unit JhDefaultPrinter;

{
  Name: TJhDefaultPrinter
  Author: Jim Hale Jr. (jhh09@yahoo.com)
  Date: 10/1/00

  TJhDefaultPrinter is a nonvisual component that allows you to change the
  default printer in Windows 95, 98, ME, NT, and 2000. It also gives easy access
  to the names of the installed printers.
}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs;

const
  MAXPRINTERBUFFER = 8000;
  // max length the names of all the installed printers concatenated together
  // could be

  MAXPRINTERNAME = 500;
  // max length a single printer name could be

  MAXPRINTERINFO = 50;
  // max length a printer's spooler and port name concatenated together could be

type
  TWhichOS = (win95, win98, winNT3, winNT4, win2k);
  TPrinterBuffer = array[0..MAXPRINTERBUFFER - 1] of char;

  TJhDefaultPrinter = class(TComponent)
  private
    xwhichOS: TWhichOS; // which OS is the user using
    xprinterNames: TStringList; // holds names of all installed printers
    xdefaultPrinter: integer;  // index in xprinterNames of the default printer

    procedure getOS;
    // sets xwhichOS with the user's Windows version

    procedure getPrinterNames;
    // fills xprinterNames with the names of the installed printers

    procedure getDefaultPrinterIndex;
    // finds where in xprinterNames the default printer is located and sets
    // xdefaultPrinter with that location

    procedure setDefaultWin95NT(const printerName: string);
    // sets the default printer in Win 95/98/NT to the given name

    function ParseNames(const namebuffer: TPrinterBuffer;
                        var startPos: integer): string;
    // pulls the printer names out of the null delimited list Windows provides

  public
    property printerNames: TStringlist read xprinterNames;
    // list of the printer names

    property defaultPrinterIndex: integer read xdefaultPrinter;
    // index in printerNames of the default printer, if there is no default
    // printer this returns -1

    constructor Create(AOwner: TComponent); override;

    function getDefaultPrinterName: string;
    // returns the name of the default printer, empty string if there isn't one

    function setDefaultPrinter(const position: integer): boolean; overload;
    // sets the default printer to the one whose position in the printerNames
    // list is position.  True if successful, false otherwise.

    function setDefaultPrinter(const name: string): boolean; overload;
    // sets the default printer to the one with the given name.  True if
    // successful, false otherwise.

    procedure Refresh;
    // retrieves the names of all installed printers and sets defaultPrinter
    // with the position in printerNames of the default printer

    destructor Destroy; override;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('System', [TJhDefaultPrinter]);
end;

constructor TJhDefaultPrinter.Create(AOwner: TComponent);
begin
  getOS;
  Refresh;
  inherited Create(AOwner);
end;


procedure TJhDefaultPrinter.getDefaultPrinterIndex;
var
  i: integer;
  printerName: string;
begin
  xdefaultPrinter := -1;  // in case there is no default

  // first we get the default printer name...

  // note: this code for getting the default printer name works on win 95, 98,
  // nt, and 2k.  An easier (and faster) way in win2k is to use the
  // GetDefaultPrinter api call, but that was not available in Delphi at the
  // time of this writing.

  SetLength(printerName, MAXPRINTERNAME);

  GetProfileString(PChar('Windows'), PChar('Device'), '', PChar(printerName),
                   MAXPRINTERNAME);
  Setlength(printerName, StrLen(PChar(printerName)));

  // pull the name out from the given name since the given name includes the
  // spooler and port name

  printerName := Copy(printerName, 1, Pos(',', printerName) - 1);

  // look for the position of the default printer in xprinterNames, that
  // position is our xdefaultPrinter

  for i := 0 to xprinterNames.Count - 1 do
  begin
    if xprinterNames.Strings[i] = printerName then
    begin
      xdefaultPrinter := i;
      break;
    end;
  end;
end;


procedure TJhDefaultPrinter.getOS;
var
  osInfo: TOSVersionInfo;
begin
  osInfo.dwOSVersionInfoSize := SizeOf(osInfo);
  if not GetVersionEx(osInfo) then
    raise Exception.Create('Unable to get OS Version Info');  

  with osInfo do
  begin
    if dwPlatformID = VER_PLATFORM_WIN32_WINDOWS then  // windows 95/98/ME
    begin
      if dwMinorVersion = 0 then
        xwhichOS := win95
      else
        xwhichOS := win98;
    end
    else  // dwPlatformID = VER_PLATFORM_WIN32_NT, win NT/2000 or higher
    begin
      If dwMajorVersion = 3 then
        xwhichOS := winNT3
      else if dwMajorVersion = 4 then
        xwhichOS := winNT4
      else // 5 or higher, windows 2000 and above
        xwhichOS := win2k;
    end;
  end;
end;


procedure TJhDefaultPrinter.getPrinterNames;
var
  buffer: TPrinterBuffer; // holds names of all printers concatenated together
  currPos: integer; // current position in buffer as we parse out the names
  printerName: string; // current name we're adding
begin
  xprinterNames.Free;
  xprinterNames := TStringList.Create;

  // calling GetProfileString with null for the registry key name will cause
  // Windows to concatenate the names of all the keys under the parent key
  // (passed in as the first param).  The names are delimited by the null char
  // (Chr(0)).

  // GetProfileString returns the number of chars it stored in buffer, so if
  // it returns 0 there are no printers installed on this machine.

  if GetProfileString(PChar('PrinterPorts'), nil, '',
                      buffer, MAXPRINTERBUFFER) > 0 then
  begin
    currPos := 0;

    while (true) do
    begin
      printerName := ParseNames(buffer, currPos);
      if printerName <> '' then
        xprinterNames.Add(printerName)
      else
        break;
    end;
  end;
end;


function TJhDefaultPrinter.ParseNames(const namebuffer: TPrinterBuffer;
                                    var startPos: integer): string;
var
  i, j: integer;
  str: string;
  nameLength: integer;
begin
  result := ''; // in case there is either nothing but nulls left in buffer,
                // or there were too many printer names to fit into buffer; if
                // that is the case then only the printers whose full names
                // fit inside buffer will appear inside printerNames

  if (startPos > High(namebuffer)) or (namebuffer[startPos] = Chr(0)) then
    exit;

  // search for the null char delimiter, when it's found strip out the
  // printer name.
  
  for i := startPos to High(namebuffer) do
  begin
    if namebuffer[i] = Chr(0) then
    begin
      nameLength := i - startPos;
      SetLength(str, nameLength);
      for j := 0 to nameLength - 1 do
        str[j+1] := namebuffer[startPos + j];
      result := str;
      startPos := i + 1;
      break;
    end
  end;
end;


destructor TJhDefaultPrinter.Destroy;
begin
  xprinterNames.Free;
  inherited Destroy;
end;


function TJhDefaultPrinter.getDefaultPrinterName: string;
begin
  if xdefaultPrinter <> -1 then
    result := xprinterNames.Strings[xdefaultPrinter]
  else
    result := '';
end;


function TJhDefaultPrinter.setDefaultPrinter(const name: string): boolean;
begin
{
  Ideally, the code would be as follows:

  try
    if (xwhichOS = win9598) or (whichOS = winNT) then
      setDefaultWin95NT(name)
    else // win 2k or later
    begin
      if not SetDefaultPrinter(PChar(printerName)) then
        raise Exception.Create('Unable to set default printer');
    end;

    xdefaultPrinter := xprinterNames.IndexOf(name);
    result := true;
  except
    result := false;
  end;

  Unfortunately, Delphi did not support the Win 2k api call SetDefaultPrinter
  at the time of this writing, so we'll have to use the slower Win 95/NT method
  for now...
}

  try
    setDefaultWin95NT(name);
    xdefaultPrinter := xprinterNames.IndexOf(name);
    result := true;
  except
    result := false;
  end;
end;


function TJhDefaultPrinter.setDefaultPrinter(const position: integer): boolean;
begin
  if (position > -1) and (position < xprinterNames.count) then
    result := setDefaultPrinter(xprinterNames[position])
  else
    result := false;
end;


procedure TJhDefaultPrinter.setDefaultWin95NT(const printerName: string);
var
  printerInfo: string;
begin
  SetLength(printerInfo, MAXPRINTERINFO);

  // first retrieve the printer's spooler and port names

  if GetProfileString(PChar('PrinterPorts'), PChar(printerName), '',
                      PChar(printerInfo), MAXPRINTERINFO) > 0 then
  begin
    SetLength(printerInfo, StrLen(PChar(printerInfo)));

    // the info needs to be written in the format:
    //    printerName,spoolerName,portName

    printerInfo := printerName + ',' +
                   Copy(printerInfo, 1, Pos(':,', printerInfo));

    if not WriteProfileString(PChar('Windows'), PChar('Device'),
                              PChar(printerInfo)) then
      raise Exception.Create('Unable to write new default printer info')
    else
    begin
      if xwhichOS in [win95, winNT3] then
        SendMessage(HWND_BROADCAST, WM_WININICHANGE, 0, 0)
      else
        SendMessage(HWND_BROADCAST, WM_SETTINGCHANGE, 0, 0);
    end;
  end
  else // unable to retrieve printer's spooler and port name
    raise Exception.Create('Unable to retrieve the printer info');
end;


procedure TJhDefaultPrinter.Refresh;
begin
  getPrinterNames;

  if xprinterNames.Count > 0 then
    getDefaultPrinterIndex;
end;

end.
 